::client_cache <- [];
::player_cache <- [ [], [], [], [] ];
::alive_player_cache <- [ [], [], [], [] ];

for (local i = 1; i <= MAX_PLAYERS; i++)
{
    local player = PlayerInstanceFromIndex(i);
    if (player)
    {
        local team = player.GetTeam();

        client_cache.push(player);
        if (player && team > 1)
        {
            player_cache[0].push(player);
            player_cache[team].push(player);
            if (GetPropInt(player, "m_lifeState") == 0)
            {
                alive_player_cache[0].push(player);
                alive_player_cache[team].push(player);
            }
        }
    }
}

//A valid _client_ can be a spectator. A valid _player_ can not.
::GetClients <- function()
{
    return client_cache;
}

//A valid _client_ can be a spectator. A valid _player_ can not.
::GetPlayers <- function(team = 0)
{
    return player_cache[team];
}

::GetAlivePlayers <- function(team = 0)
{
    return alive_player_cache[team];
}

::CTFPlayer.IsAlive <- function()
{
    return GetPropInt(this, "m_lifeState") == 0;
}
::CTFBot.IsAlive <- CTFPlayer.IsAlive;

//A valid _client_ can be a spectator. A valid _player_ can not.
::IsValidClient <- function(player)
{
    try { return player && player.IsValid() && player.IsPlayer(); } catch(e) { return false; }
}

::ClientCache_Join <- function(player)
{
    if (client_cache.find(player) == null)
        client_cache.push(player);
}

OnGameEvent("player_team", -1000, function(player, params)
{
    foreach(team in [0, params.oldteam])
    {
        local index = player_cache[team].find(player);
        if (index != null) player_cache[team].remove(index);
        index = alive_player_cache[team].find(player);
        if (index != null) alive_player_cache[team].remove(index);
    }

    if (params.team >= 2)
    {
        player_cache[0].push(player);
        player_cache[params.team].push(player);
    }
});

OnGameEvent("player_spawn", -1000, function(player, params)
{
    if (GetPropInt(player, "m_lifeState") == 0)
        foreach(team in [0, player.GetTeam()])
            if (alive_player_cache[team].find(player) == null)
                alive_player_cache[team].push(player);
});

OnGameEvent("player_death_post", 1000, function(player, params)
{
    local index;
    foreach(team in [0, player.GetTeam()])
        if ((index = alive_player_cache[team].find(player)) != null)
            alive_player_cache[team].remove(index);
});

OnGameEvent("player_disconnect", 1000, function(player, params)
{
    local index = client_cache.find(player);
    if (index != null) client_cache.remove(index);

    foreach(team in [0, player.GetTeam()])
    {
        index = player_cache[team].find(player);
        if (index != null) player_cache[team].remove(index);
        index = alive_player_cache[team].find(player);
        if (index != null) alive_player_cache[team].remove(index);
    }
});