//!CompilePal::IncludeDirectory("resource/arena_vscript")
//!CompilePal::IncludeDirectory("scripts/vscripts/arena_vscript")

IncludeScript("arena_vscript/_lizardlib/_lizardlib.nut");
IncludeConfig("arena_vscript/config.nut");
try { IncludeScript("arena_vscript_addons/preload.nut"); } catch(e) { }
IncludeScript("arena_vscript/debug.nut");

EntFireByHandle(tf_gamerules, "SetRedTeamRespawnWaveTime", "99999", 0, null, null);
EntFireByHandle(tf_gamerules, "SetBlueTeamRespawnWaveTime", "99999", 0, null, null);
game_forcerespawn <- SpawnEntityFromTable("game_forcerespawn", {});
tf_objective_resource <- Entities.FindByClassname(null, "tf_objective_resource");

roundCounter <- -1;
isMiniRoundSetup <- true;
isMiniRoundOver <- false;
setupEndTS <- 0;
wasWaitingForPlayers <- false;
respawnRecursionFlag <- false;
hasRespawned <- {};

//================================================
//Mini-Round Logic
//================================================

function ResetNewMiniRound()
{
    if (GetPropInt(tf_gamerules, "m_iRoundState") == GR_STATE_TEAM_WIN)
        return;

    RunWithDelay(SETUP_LENGTH, BeginMiniRound);

    roundCounter++;
    isMiniRoundOver = false;
    isMiniRoundSetup = true;
    wasWaitingForPlayers = false;
    overtimeTimer = null;
    firstBloodUsed = false;
    isLastMannStandingActive = false;
    hasRespawned = {};

    foreach (control_point in control_points)
    {
        control_point.AcceptInput("SetOwner", "0", control_point, control_point);
        control_point.AcceptInput("SetLocked", "1", null, null);
    }

    EntFireByHandle(game_forcerespawn, "ForceRespawn", "", 0, null, null);
    DoEntFire("tf_dropped_weapon", "Kill", "", 0, null, null);
    DoEntFire("func_respawnroom", "Enable", "", 0, null, null);

    DoEntFire("arena_reset*", "Trigger", "", 0, null, null);
    DoEntFire("arena_start*", "Close", "", 0, null, null);

    DisplayCountdown();
    setupEndTS = Time() + SETUP_LENGTH;
    FireCustomEvent("arena_miniround_reset", { round = roundCounter });
}

function DisplayCountdown()
{
    local team_round_timer = SpawnEntityFromTable("team_round_timer", {
        auto_countdown = 0,
        max_length = 0,
        timer_length = SETUP_LENGTH,
        reset_time = 1,
        setup_length = 0,
        show_in_hud = 1,
        show_time_remaining = 1,
        start_paused = 0,
        StartDisabled = 0
    });
    SetPropBool(team_round_timer, "m_bForcePurgeFixedupStrings", true);
    EntFireByHandle(team_round_timer, "Resume", "", 0, null, null);
    EntFireByHandle(team_round_timer, "Kill", "", SETUP_LENGTH, null, null);

    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins10Seconds", SETUP_LENGTH - 11, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins5Seconds", SETUP_LENGTH - 6, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins4Seconds", SETUP_LENGTH - 5, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins3Seconds", SETUP_LENGTH - 4, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins2Seconds", SETUP_LENGTH - 3, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.RoundBegins1Seconds", SETUP_LENGTH - 2, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Announcer.AM_RoundStartRandom", SETUP_LENGTH, null, null);
    EntFireByHandle(tf_gamerules, "PlayVO", "Ambient.Siren", SETUP_LENGTH, null, null);
}

function BeginMiniRound()
{
    game_forcerespawn.AcceptInput("ForceTeamRespawn", "2", null, null);
    game_forcerespawn.AcceptInput("ForceTeamRespawn", "3", null, null);

    isMiniRoundSetup = false;

    deathCountdown = CalculateDeathQuota(GetPlayers().len());
    if (UNLOCK_ON_DEATH && deathCountdown <= 0)
        foreach (control_point in control_points)
            EntFireByHandle(control_point, "SetLocked", "0", 0, null, null);

    if (UNLOCK_TIMER_TIME > 0.1)
        foreach (control_point in control_points)
            EntFireByHandle(control_point, "SetUnlockTime", UNLOCK_TIMER_TIME.tostring(), 0, null, null);

    DoEntFire("func_respawnroom", "Disable", "", 0, null, null);

    DoEntFire("arena_start*", "Trigger", "", 0, null, null);
    DoEntFire("arena_start*", "Open", "", 0, null, null);

    FireCustomEvent("arena_miniround_start", { round = roundCounter });
}

function EndMiniRound(winnerTeam)
{
    if (isMiniRoundOver)
        return;
    isMiniRoundOver = true;

    RunWithDelay(HUMILIATION_LENGTH, ResetNewMiniRound);

    local loserTeam = winnerTeam == TF_TEAM_RED ? TF_TEAM_BLUE : winnerTeam == TF_TEAM_BLUE ? TF_TEAM_RED: 0;

    foreach(player in GetPlayers(winnerTeam))
        player.AddCond(TF_COND_CRITBOOSTED_PUMPKIN);

    foreach(player in GetPlayers(loserTeam))
    {
        player.StunPlayer(HUMILIATION_LENGTH + 1, 1, 96, null);
        player.StopSound("Halloween.PlayerScream");
    }

    for (local ent = null; ent = Entities.FindByClassname(ent, "obj_sentrygun");)
    {
        SetPropBool(ent, "m_bForcePurgeFixedupStrings", true);
        if (ent.GetTeam() != winnerTeam)
            EntFireByHandle(ent, "Disable", "", 0, null, null);
    }

    DoEntFire("arena_end_any*", "Trigger", "", 0, null, null);
    if (winnerTeam == TF_TEAM_RED)
    {
        DoEntFire("arena_end_red*", "Trigger", "", 0, null, null);
        EntFireByHandle(pd_logic, "scoreredpoints", "", 1, null, null);
    }
    else if (winnerTeam == TF_TEAM_BLUE)
    {
        DoEntFire("arena_end_blu*", "Trigger", "", 0, null, null);
        EntFireByHandle(pd_logic, "scorebluepoints", "", 1, null, null);
    }
    else
        DoEntFire("arena_end_stalemate*", "Trigger", "", 0, null, null);

    EntFireByHandle(tf_gamerules, "PlayVO", "Hud.EndRoundScored", 0, null, null);
    DisplayVictoryMessage(winnerTeam);
    FireCustomEvent("arena_miniround_end", { team = winnerTeam });
}

//================================================
//Player-Related Arena Logic
//================================================

function CheckAlivePlayers()
{
    if (IsInWaitingForPlayers() || isMiniRoundOver)
        return;

    local redTotal = GetPlayers(TF_TEAM_RED).len();
    local bluTotal = GetPlayers(TF_TEAM_BLUE).len();

    if (isMiniRoundSetup)
    {
        deathCountdown = CalculateDeathQuota(redTotal + bluTotal);
        return;
    }

    if (redTotal < 1 || bluTotal < 1)
    {
        wasWaitingForPlayers = true;
        return;
    }
    else if (wasWaitingForPlayers)
    {
        ResetNewMiniRound();
        return;
    }

    local redAlive = GetAlivePlayers(TF_TEAM_RED).len();
    local bluAlive = GetAlivePlayers(TF_TEAM_BLUE).len();

    if (redAlive == 0 && bluAlive == 0)
        EndMiniRound(TEAM_UNASSIGNED);
    else if (redAlive == 0)
        EndMiniRound(TF_TEAM_BLUE);
    else if (bluAlive == 0)
        EndMiniRound(TF_TEAM_RED);
}
AddTimer(0.1, CheckAlivePlayers);

function OnPlayerDeathPost(player)
{
    if (!isMiniRoundSetup)
    {
        UpdatePlayerCounters();
        OnTickEnd(CheckAlivePlayers);
    }
    else if (!(player in hasRespawned))
    {
        hasRespawned[player] <- true;
        RunWithDelay(1, function(player)
        {
            if (isMiniRoundSetup)
                player.ForceRespawn();
        }, player);
    }
}
OnGameEvent("player_death_post", 0, OnPlayerDeathPost);

if (HOLD_PLAYERS_DURING_SETUP)
{
    function HoldPlayerDuringSetup(player)
    {
        player.AddCustomAttribute("move speed penalty", 0.01, setupEndTS - Time());
        local tauntSpeed = [0];
        AddTimer(-1, function(setupEndTS)
        {
            local tauntSpeedL = self.GetCurrentTauntMoveSpeed();
            if (tauntSpeedL > 0)
            {
                tauntSpeed[0] = tauntSpeedL;
                self.SetCurrentTauntMoveSpeed(0);
            }
            if (Time() >= setupEndTS)
            {
                self.SetCurrentTauntMoveSpeed(tauntSpeed[0]);
                return TIMER_STOP;
            }
        }, setupEndTS, player)
    }

    OnGameEvent("player_spawn", 0, function(player, params)
    {
        if (!respawnRecursionFlag && isMiniRoundSetup)
            OnTickEnd(HoldPlayerDuringSetup, player);
    });
}

function PreventMidRoundSpawn(player, params)
{
    if (!respawnRecursionFlag && !isMiniRoundSetup)
    {
        respawnRecursionFlag = true;
        SetPropInt(player, "m_iObserverLastMode", 5);
        local team = player.GetTeam();
        SetPropInt(player, "m_iTeamNum", 1);
        player.DispatchSpawn();
        SetPropInt(player, "m_iTeamNum", team);
        respawnRecursionFlag = false;
    }
}
OnGameEvent("player_spawn", -1001, PreventMidRoundSpawn);

function PreventSuicideOnClassChange(player)
{
    local weapon = player.GetActiveWeapon();
    player.AddCond(TF_COND_HALLOWEEN_GHOST_MODE);
    OnTickEnd(function()
    {
        player.RemoveCond(TF_COND_HALLOWEEN_GHOST_MODE);
        player.Weapon_Switch(weapon);
    })
}
OnGameEvent("player_changeclass", 0, PreventSuicideOnClassChange);

//================================================
//Point Capture Logic
//================================================

overtimeTimer <- null;
control_points <- [];

function CollectControlPoints()
{
    for (local ent = null; ent = Entities.FindByClassname(ent, "team_control_point");)
        control_points.push(ent);
}
OnTickEnd(CollectControlPoints);

function CheckPointStatus()
{
    if (isMiniRoundOver)
        return TIMER_STOP;

    local team = control_points[0].GetTeam();
    foreach (control_point in control_points)
        if (control_point.GetTeam() != team)
            return;
    for (local i = 0; i < 8; i++)
        if (GetPropFloatArray(tf_objective_resource, "m_flCapPercentages", i) - 0 > Epsilon)
        {
            if (!overtimeTimer)
                overtimeTimer = AddTimer(-1, CheckPointStatus);
            return;
        }
    EndMiniRound(team);
}

function AddPointInput_CheckPointStatus()
{
    foreach (control_point in control_points)
    {
        EntityOutputs.AddOutput(control_point, "OnCapTeam1", "!self", "RunScriptCode", "main_script.CheckPointStatus()", 0, -1);
        EntityOutputs.AddOutput(control_point, "OnCapTeam2", "!self", "RunScriptCode", "main_script.CheckPointStatus()", 0, -1);
    }
}
OnTickEnd(AddPointInput_CheckPointStatus);

//================================================
//HUD
//================================================

SetPropInt(tf_gamerules, "m_nHudType", 2);
worldspawn.SetTeam(TF_TEAM_BLUE);

pd_logic <- SpawnEntityFromTable("tf_logic_player_destruction", {
    min_points = VICTORY_SCORE,
    finale_length = 0,
    res_file = HUD_PATH
});
pd_logic.AcceptInput("SetPointsOnPlayerDeath", "0", null, null);
pd_logic.AcceptInput("EnableMaxScoreUpdating", "0", null, null);
pd_logic.AcceptInput("DisableMaxScoreUpdating", "0", null, null);

function SpawnEscrowFlags()
{
    escrow_red <- SpawnEntityFromTable("item_teamflag", {
        PointValue = 1,
        flag_model = "models/empty.mdl",
        origin = Vector(-9999, -9999, -9999),
        GameType = 6
    });
    SetPropInt(escrow_red, "m_nFlagStatus", 1);

    escrow_blue <- SpawnEntityFromTable("item_teamflag", {
        PointValue = 1,
        flag_model = "models/empty.mdl",
        origin = Vector(-9999, -999, -999),
        GameType = 6
    });
    SetPropEntity(escrow_blue, "m_hPrevOwner", worldspawn);
    SetPropInt(escrow_blue, "m_nFlagStatus", 1);

    UpdatePlayerCounters();
}
OnTickEnd(SpawnEscrowFlags);

function UpdatePlayerCounters()
{
    if (UNLOCK_ON_DEATH)
        EntFireByHandle(pd_logic, "setcountdowntimer", deathCountdown <= 0 ? "0" : (deathCountdown + 1.2).tostring(), 0, null, null);

    foreach (player in GetPlayers(TF_TEAM_RED))
    {
        SetPropEntity(escrow_red, "m_hPrevOwner", player);
        break;
    }
    SetPropInt(escrow_red, "m_nPointValue", GetAlivePlayers(TF_TEAM_RED).len());
    SetPropInt(escrow_blue, "m_nPointValue", GetAlivePlayers(TF_TEAM_BLUE).len());
}
AddTimer(0.3, UpdatePlayerCounters);

function DisplayVictoryMessage(winnerTeam)
{
    local game_text = SpawnEntityFromTable("game_text", {
        color = (winnerTeam == TF_TEAM_RED ? "255 25 25" : winnerTeam == TF_TEAM_BLUE ? "25 25 255" : "255 255 255"),
        color2 = "0 0 0",
        effect = 0,
        fadein = 0.2,
        fadeout = 0.2,
        fxtime = 0,
        holdtime = 5,
        message = (winnerTeam == TF_TEAM_RED ? "RED WIN" : winnerTeam == TF_TEAM_BLUE ? "BLU WIN" : "STALEMATE"),
        spawnflags = 1,
        channel = 1,
        x = -1,
        y = 0.7
    });
    SetPropBool(game_text, "m_bForcePurgeFixedupStrings", true);
    EntFireByHandle(game_text, "Display", "", 0, null, null);
    EntFireByHandle(game_text, "Kill", "", 1, null, null);
}

//================================================
//First Blood
//================================================

firstBloodUsed <- false;

if (FIRST_BLOOD_DURATION >= 0.1)
{
    function OnDeath_FirstBlood(params)
    {
        if (firstBloodUsed)
            return;
        local attacker = GetPlayerFromUserID(params.attacker);
        if (!IsValidClient(attacker))
            return;

        attacker.AddCondEx(TF_COND_CRITBOOSTED_FIRST_BLOOD, FIRST_BLOOD_DURATION, attacker);
        firstBloodUsed = true;
    }
    OnGameEvent("player_death", 0, OnDeath_FirstBlood);
}

//================================================
//Last Mann Standing
//================================================

isLastMannStandingActive <- false;

if (LAST_MANN_STANDING_ENABLED)
{
    function CheckLastMannStanding()
    {
        if (IsInWaitingForPlayers()
            || isMiniRoundSetup
            || GetPlayers().len() < LAST_MANN_STANDING_PLAYER_TOTAL_COUNT
            || isLastMannStandingActive)
            return;

        if (GetAlivePlayers(TF_TEAM_RED).len() == 1 || GetAlivePlayers(TF_TEAM_BLUE).len() == 1)
        {
            isLastMannStandingActive = true;
            AddTimer(0.1, TickLastMannStanding);
        }
    }
    OnGameEvent("player_death_next", 0, CheckLastMannStanding);
    OnGameEvent("player_disconnect_next", 0, CheckLastMannStanding);
    OnSelfEvent("player_team_next", 0, CheckLastMannStanding);

    function TickLastMannStanding()
    {
        if (isMiniRoundSetup)
        {
            isLastMannStandingActive = false;
            return TIMER_STOP;
        }

        foreach(team in [TF_TEAM_RED, TF_TEAM_BLUE])
        {
            local alivePlayers = GetAlivePlayers(team);
            if (alivePlayers.len() <= LAST_MANN_STANDING_PLAYERS_LEFT)
            {
                foreach (player in alivePlayers)
                    if (player && !player.InCond(TF_COND_STEALTHED) && !player.InCond(TF_COND_DISGUISED))
                        foreach (buff in LAST_MANN_STANDING_BUFFS)
                            player.AddCondEx(buff, 0.2, null);
            }
        }
    }
}

//================================================
//On-Capture Respawn
//================================================

deathCountdown <- 0;

if (UNLOCK_ON_DEATH)
{
    function OnDeath_DeathCounter(player)
    {
        if (deathCountdown <= 0)
            return;
        deathCountdown--;

        if (deathCountdown <= 0)
        {
            DoEntFire("arena_unlock*", "Trigger", "", 0, null, null);
            foreach (control_point in control_points)
                EntFireByHandle(control_point, "SetLocked", "0", 0, null, null);
        }
    }
    OnGameEvent("player_death", 0, OnDeath_DeathCounter);
}

//================================================
//On-Capture Respawn
//================================================

if (ON_CAPTURE_RESPAWN)
{
    function CollectRespawnLocations()
    {
        respawn_locations <- {};
        foreach (control_point in control_points)
        {
            local array = [];
            respawn_locations[control_point] <- array;
            local respawnName = control_point.GetName() + "_respawn";
            for (local ent = null; ent = Entities.FindByName(ent, respawnName);)
                array.push([ent.GetOrigin(), ent.GetAbsAngles()]);
        }
    }
    OnTickEnd(CollectRespawnLocations);

    function RespawnTeam(control_point, team)
    {
        respawnRecursionFlag = true;
        local locations = respawn_locations[control_point];
        local i = 0;
        foreach(player in GetPlayers(team))
            if (!player.IsAlive())
            {
                local location = locations[i++];
                if (i >= locations.len())
                    i = 0;

                player.ForceRespawn();
                player.AddCondEx(TF_COND_INVULNERABLE, 1.2, null);
                player.Teleport(true, location[0], true, location[1], true, Vector(0, 0, 0));
                EmitSoundOnClient("Halloween.Merasmus_TP_In", player);
            }
        respawnRecursionFlag = false;
    }

    function AddPointInput_RespawnTeam()
    {
        foreach (control_point in control_points)
        {
            EntityOutputs.AddOutput(control_point, "OnCapTeam1", "!self", "RunScriptCode", "main_script.RespawnTeam(self, TF_TEAM_RED)", 0, -1);
            EntityOutputs.AddOutput(control_point, "OnCapTeam2", "!self", "RunScriptCode", "main_script.RespawnTeam(self, TF_TEAM_BLUE)", 0, -1);
        }
    }
    OnTickEnd(AddPointInput_RespawnTeam);
}

//===================================================================
//PostLoad
//===================================================================

OnTickEnd(ResetNewMiniRound);

try { IncludeScript("arena_vscript_addons/postload.nut"); } catch(e) { }

__CollectGameEventCallbacks(this);